"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Exceptions;
(function (Exceptions) {
    class OpenFinTSClientException extends Error {
        constructor(message) {
            super(message);
            Object.setPrototypeOf(this, OpenFinTSClientException.prototype);
        }
        toString() {
            return this.message ? this.message : 'OpenFinTSClientException';
        }
    }
    Exceptions.OpenFinTSClientException = OpenFinTSClientException;
    class GVNotSupportedByKI extends OpenFinTSClientException {
        constructor(gvType, avail) {
            super('There is no version of ' + gvType + ' which is supported by both, the client and the server.');
            this.spVers = avail ? [] : Object.keys(avail);
            Object.setPrototypeOf(this, GVNotSupportedByKI.prototype);
        }
    }
    Exceptions.GVNotSupportedByKI = GVNotSupportedByKI;
    class MalformedMessageFormat extends OpenFinTSClientException {
        constructor(message) {
            super('MalformedMessageFormat: ' + message);
            Object.setPrototypeOf(this, MalformedMessageFormat.prototype);
        }
    }
    Exceptions.MalformedMessageFormat = MalformedMessageFormat;
    class OrderFailedException extends OpenFinTSClientException {
        constructor(msg) {
            super('Failed to perform Order, got error Message from Server.:' + msg.getEl(3));
            this.msgDetail = msg;
            Object.setPrototypeOf(this, OrderFailedException.prototype);
        }
    }
    Exceptions.OrderFailedException = OrderFailedException;
    class InternalError extends OpenFinTSClientException {
        constructor(message) {
            super(message);
            Object.setPrototypeOf(this, InternalError.prototype);
        }
    }
    Exceptions.InternalError = InternalError;
    class GVFailedAtKI extends OpenFinTSClientException {
        constructor(msg) {
            super('GVFailed because Msg: ' + msg.data[0].data + ' - ' + msg.data[2].data);
            this.data = msg;
            Object.setPrototypeOf(this, GVFailedAtKI.prototype);
        }
    }
    Exceptions.GVFailedAtKI = GVFailedAtKI;
    class ConnectionFailedException extends OpenFinTSClientException {
        constructor(hostname, port, path) {
            super(hostname);
            this.host = hostname;
            Object.setPrototypeOf(this, ConnectionFailedException.prototype);
        }
        toString() {
            return 'Connection to ' + this.host + ' failed.';
        }
    }
    Exceptions.ConnectionFailedException = ConnectionFailedException;
    class MissingBankConnectionDataException extends OpenFinTSClientException {
        constructor(blz) {
            super(blz);
            this.blz = blz;
            Object.setPrototypeOf(this, MissingBankConnectionDataException.prototype);
        }
        toString() {
            return 'No connection Url in Bankenliste found to connect to blz: ' + this.blz + '.';
        }
    }
    Exceptions.MissingBankConnectionDataException = MissingBankConnectionDataException;
    class OutofSequenceMessageException extends OpenFinTSClientException {
        constructor(blz) {
            super(blz);
            this.blz = blz;
            Object.setPrototypeOf(this, OutofSequenceMessageException.prototype);
        }
        toString() {
            return 'You have to ensure that only one message at a time is send to the server, use libraries like async or promisses. You can send a new message as soon as the callback returns.';
        }
    }
    Exceptions.OutofSequenceMessageException = OutofSequenceMessageException;
})(Exceptions = exports.Exceptions || (exports.Exceptions = {}));
