"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ByteVal_1 = require("./ByteVal");
const DatenElementGruppe_1 = require("./DatenElementGruppe");
const Figure_1 = require("./Figure");
const Balance_1 = require("./Balance");
const Segment_1 = require("./Segment");
const TransactionType_1 = require("./TransactionType");
class Helper {
    static checkMsgsWithBelongToForId(msg, bez, id) {
        const array = msg.getSegmentsByNameAndReference('HIRMS', bez);
        if (array.length <= 0) {
            return null;
        }
        for (let i = 0; i !== array.length; i += 1) {
            let result = null;
            array[0].store.data.forEach(el => {
                if (el.data.getEl(1) === id) {
                    result = el.data;
                }
            });
            return result;
        }
        return null;
    }
    static getNrWithLeadingNulls(nr, len) {
        const stxt = nr + '';
        let ltxt = '';
        const neu = len - stxt.length;
        for (let i = 0; i !== neu; i += 1) {
            ltxt += '0';
        }
        ltxt += stxt;
        return ltxt;
    }
    static newSegFromArray(name, vers, ar) {
        const seg = new Segment_1.Segment();
        seg.init(name, 0, vers, 0);
        for (let i = 0; i !== ar.length; i += 1) {
            if (ar[i] instanceof Array) {
                const neu = new DatenElementGruppe_1.DatenElementGruppe();
                for (let j = 0; j !== ar[i].length; j += 1) {
                    if (ar[i][j] instanceof ByteVal_1.ByteVal) {
                        neu.addDEbin(ar[i][j].data);
                    }
                    else {
                        neu.addDE(ar[i][j]);
                    }
                }
                seg.store.addDEG(neu);
            }
            else if (ar[i] instanceof ByteVal_1.ByteVal) {
                seg.store.addDEbin(ar[i].data);
            }
            else {
                // normales datenelement
                seg.store.addDE(ar[i]);
            }
        }
        return seg;
    }
    static convertIntoArray(deOrDeg) {
        if (deOrDeg instanceof DatenElementGruppe_1.DatenElementGruppe) {
            return deOrDeg.data;
        }
        return [deOrDeg];
    }
    static convertDateToDFormat(date) {
        const yyyy = date.getFullYear() + '';
        const mm = ((date.getMonth() + 1) <= 9) ? ('0' + (date.getMonth() + 1)) : ((date.getMonth() + 1) + '');
        const dd = (date.getDate() <= 9) ? ('0' + date.getDate()) : (date.getDate() + '');
        return yyyy + mm + dd;
    }
    static convertDateToTFormat(date) {
        const hh = ((date.getHours() <= 9) ? '0' : '') + date.getHours();
        const mm = ((date.getMinutes() <= 9) ? '0' : '') + date.getMinutes();
        const ss = ((date.getSeconds() <= 9) ? '0' : '') + date.getSeconds();
        return hh + mm + ss;
    }
    static convertFromToJSText(ftxt) {
        let jstxt = '';
        const re = /\?([^\?])/g;
        jstxt = ftxt.replace(re, '$1');
        return jstxt;
    }
    static convertJSTextTo(jstxt) {
        let ftxt = '';
        const re = /([:\+\?'\@])/g;
        ftxt = jstxt.replace(re, '?$&');
        return ftxt;
    }
    static byte(data) {
        return new ByteVal_1.ByteVal(data);
    }
    static getSaldo(seg, nr, hbciVer3) {
        if (!seg) {
            return null;
        }
        try {
            const base = seg.getEl(nr).data;
            const result = new Balance_1.Balance();
            result.transactionType = base.getEl(1) === 'C' ? TransactionType_1.TransactionType.CREDIT : TransactionType_1.TransactionType.DEBIT;
            result.currency = hbciVer3 ? 'EUR' : base.getEl(3);
            result.value = parseFloat(base.getEl(2).replace(',', '.'));
            result.entryDate = this.getJSDateFromSeg(base, hbciVer3 ? 3 : 4, hbciVer3 ? 4 : 5);
            return result;
        }
        catch (ee) {
            return null;
        }
    }
    static getBetrag(seg, nr) {
        if (!seg) {
            return null;
        }
        try {
            const base = seg.getEl(nr).data;
            const result = new Figure_1.Figure();
            result.currency = base.getEl(2);
            result.value = parseFloat(base.getEl(1).replace(',', '.'));
            return result;
        }
        catch (ee) {
            return null;
        }
    }
    static getJSDateFromSegTSP(seg, nr) {
        try {
            const base = seg.getEl(nr).data;
            return this.getJSDateFromSeg(base, 1, 2);
        }
        catch (e) {
            return null;
        }
    }
    static getJSDateFromSeg(seg, dateNr, timeNr) {
        if (!seg) {
            return null;
        }
        try {
            const date = seg.getEl(dateNr);
            let time = '000000';
            try {
                if (timeNr)
                    time = seg.getEl(timeNr);
            }
            catch (eee) {
                // do nothing
            }
            const result = new Date();
            result.setTime(0);
            result.setFullYear(parseInt(date.substr(0, 4), 10));
            result.setMonth(parseInt(date.substr(4, 2), 10) - 1);
            result.setDate(parseInt(date.substr(6, 2), 10));
            result.setHours(parseInt(time.substr(0, 2), 10));
            result.setMinutes(parseInt(time.substr(2, 2), 10));
            result.setSeconds(parseInt(time.substr(4, 2), 10));
            return result;
        }
        catch (ee) {
            return null;
        }
    }
    static escapeUserString(str) {
        // escapes special characters with a '?'
        // use this when forwarding user defined input (such as username/password) to a server
        //
        // SOURCE: http://linuxwiki.de/HBCI/F%C3%BCrEntwickler
        // TODO: find better/official source
        return str.replace(/[?+:]/g, '?$&');
    }
}
Helper.newSegFromArrayWithBez = function (name, vers, bez, ar) {
    const seg = this.newSegFromArray(name, vers, ar);
    seg.referencedSegment = bez;
    return seg;
};
exports.Helper = Helper;
