"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Helper_1 = require("./Helper");
const NULL_1 = require("./NULL");
const Parser_1 = require("./Parser");
const ReturnCode_1 = require("./ReturnCode");
const Segment_1 = require("./Segment");
const SegmentName_1 = require("./SegmentName");
class Nachricht {
    constructor(protoVersion) {
        this.protoVersion = protoVersion;
        this.segments = [];
        this.segmentsCounter = 0;
        this.signIt = null;
        this.hnvsk = null;
        this.messageNumber = 0;
        this.sign = function (signInfo) {
            this.signIt = signInfo;
        };
    }
    init(dialogId, ongoingNumber, blz, kundenId) {
        // this is called wenn es ein outgoing message ist
        this.messageNumber = ongoingNumber;
        const seg = new Segment_1.Segment();
        seg.init(SegmentName_1.SegmentName.MESSAGE_HEADER, 1, '3', 0);
        this.addSeg(seg);
        seg.store.addDE(Helper_1.Helper.getNrWithLeadingNulls(0, 12)); // Länge
        seg.store.addDE(this.protoVersion + ''); // Version
        seg.store.addDE(dialogId); // Dialog-ID, bei 0 beginnend wird von KI bekannt gegeben
        seg.store.addDE(this.messageNumber); // Nachrichten-Nr. streng monoton von 1 ab steigen
        if (this.signIt) { // NUr für das Pin/Tan Verfahren 1 Schritt!
            // Infos hierzu: http://www.hbci-zka.de/dokumente/spezifikation_deutsch/fintsv3/FinTS_3.0_Security_Sicherheitsverfahren_HBCI_Rel_20130718_final_version.pdf Punkt B5.1
            // http://www.hbci-zka.de/dokumente/spezifikation_deutsch/fintsv3/FinTS_3.0_Security_Sicherheitsverfahren_PINTAN_Rel_20101027_final_version.pdf B8.4
            // Sicherheitsprofil ["PIN",1] = PIN und 1 Schrittverfahren
            // Sicherheitsfunktion: 999 - 1 SChrittverfahren / 2Schritt siehe BPD
            // Sicherheitskontrollreferenz: 1 // Muss mit Signaturabschluss übereinstimmen
            // Bereich der Sicherheitsapplikation,kodiert: 1 // 1: Signaturkopf und HBCI-Nutzdaten (SHM)
            // Rolle des Sicherheitslieferanten,kodiert: 1 // 1: Der Unterzeichner ist Herausgeber der signierten Nachricht, z.B. Erfasser oder Erstsignatur (ISS)
            // Sicherheitsidentifikation, Details: [1,null,0]
            // 		Bezeichner Sicherheitspartei	1		1: Message Sender (MS), wenn ein Kunde etwas an sein Kreditinstitut sendet
            // 		CID nur Chipkarte				null
            // 		Id der Partei nur Software		0		Code, welcher die (Kommunikations-)Partei identifiziert. Dieses Feld muss eine gültige, zuvor vom Banksystem angeforderte Kundensystem-ID enthalten (analog zum RSA-Verfahren). Dies gilt auch fürZweit-und Drittsignaturen.
            // 			beim Erstmal noch 0, dann auf Antwort von Bank in HISYN warten und das verwenden!
            // 	Sicherheitsreferenznummer: 1 Verhinderung der Doppeleinreichung Bei softwarebasierten Verfahren wird die Sicherheitsreferenznummer auf Basis des DE Kundensystem-ID und des DE Benutzerkennung der DEG Schlüsselnamen verwaltet.
            // 							bei Pin/Tan kann das auch einfach bei 1 beibehalten werden :), sonst müsste man das aber eigtl. incrementieren
            // 	Sicherheitsdatum und -uhrzeit [1,"20141210","003515"], 1: Bedeutung = Sicherheitszeitstempel (STS)
            // 	Hashalgorithmus: [1,999,1]
            // 		Verwendung des Hashalgorithmus,kodiert	1: Owner Hashing (OHA) (nur)
            // 		Hashalgorithmus,kodiert					999: Gegenseitig vereinbart (ZZZ); hier: RIPEMD-160 ( gibt noch andere Werte 1-6 vorallem SHAxxx
            // 		Bezeichner für Hashalgorithmusparameter	1: IVC (Initialization value, clear text)
            // 	Signaturalgorithmus: [6,10,16]
            // 		Verwendung des Signaturalgorithmus, kodiert 6: Owner Signing (OSG)
            // 		10: RSA-Algorithmus (bei RAH und RDH)
            // 		Operationsmodus, kodiert	16:	ISO 9796-1 (bei RDH)
            // 	Schlüsselname	[280,blz,kunden_id,"S",0,0]
            // 		Kreditinstitutskennung	280,blz
            // 		Benutzerkennung 		kunden_id
            // 		Schlüsselart			S	S: Signierschlüsse
            // 		Schlüsselnummer			0
            // 		Schlüsselversion		0
            const signatureId = (this.signIt.sysId + '') === '0' ? 1 : this.signIt.sigId;
            this.signIt.blz = blz;
            this.signIt.kundenId = kundenId;
            if (this.protoVersion === 300) {
                this.signIt.server === undefined
                    ? this.addSeg(Helper_1.Helper.newSegFromArray('HNSHK', 4, [
                        ['PIN', this.signIt.pinVersion === '999' ? 1 : 2], this.signIt.pinVersion, 1, 1, 1, [1, NULL_1.NULL, this.signIt.sysId], signatureId, [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                        [1, 999, 1],
                        [6, 10, 16],
                        [280, blz, kundenId, 'S', 0, 0],
                    ]))
                    : this.addSeg(Helper_1.Helper.newSegFromArray('HNSHK', 4, [
                        ['PIN', this.signIt.pinVersion === '999' ? 1 : 2], this.signIt.pinVersion, 1, 1, 1, [2, NULL_1.NULL, this.signIt.sysId], signatureId, [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                        [1, 999, 1],
                        [6, 10, 16],
                        [280, blz, kundenId, 'S', 0, 0],
                    ]));
            }
            else {
                this.signIt.server === undefined
                    ? this.addSeg(Helper_1.Helper.newSegFromArray('HNSHK', 3, [this.signIt.pinVersion, 1, 1, 1, [1, NULL_1.NULL, this.signIt.sysId], signatureId, [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                        [1, 999, 1],
                        [6, 10, 16],
                        [280, blz, kundenId, 'S', 0, 0],
                    ]))
                    : this.addSeg(Helper_1.Helper.newSegFromArray('HNSHK', 3, [this.signIt.pinVersion, 1, 1, 1, [2, NULL_1.NULL, this.signIt.sysId], signatureId, [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                        [1, 999, 1],
                        [6, 10, 16],
                        [280, blz, kundenId, 'S', 0, 0],
                    ]));
            }
        }
    }
    parse(txt) {
        const parser = new Parser_1.Parser(txt);
        while (parser.hasNext()) {
            const segm = new Segment_1.Segment();
            segm.parse(parser);
            this.segments.push(segm);
            parser.nextPos();
        }
        // prüfen ob verschlüsselt war
        if (this.segments.length === 4 && this.segments[1].name === 'HNVSK' && this.segments[2].name === 'HNVSD') {
            const first = this.segments[0];
            this.hnvsk = this.segments[1];
            const segHNVSD = this.segments[2];
            const last = this.segments[3];
            // Neue Segmente hinzufügen
            this.segments = [];
            this.segments.push(first);
            if ((this.hnvsk.version === '3' && this.hnvsk.getEl(1).data.getEl(1) === 'PIN') || (this.hnvsk.version === '2' && this.hnvsk.getElString(1) === '998')) {
                const parser2 = new Parser_1.Parser(segHNVSD.getEl(1).data);
                while (parser2.hasNext()) {
                    const segm2 = new Segment_1.Segment();
                    segm2.parse(parser2);
                    this.segments.push(segm2);
                    parser2.nextPos();
                }
            }
            else {
                throw new Parser_1.ParseError('Msg', 'Nicht unterstützte Verschlüsselungsmethode!', 0);
            }
            this.segments.push(last);
        }
    }
    transformForSend() {
        let top = this.segments[0].transformForSend();
        let body = '';
        // Signatur abschluss
        if (this.signIt) {
            // Signaturabschluss
            // Sicherheitskontrollreferenz 1 muss mit signaturkopf übereinstimmen
            // Validierungsresultat null, bleibt bei PinTan leer
            // Benutzerdefinierte Signatur [Pin,Tan], die Tan nur dann wenn durch den Geschäftsvorfall erforderlich
            if (this.signIt.server === undefined) {
                if (this.signIt.tan === NULL_1.NULL) {
                    this.addSeg(Helper_1.Helper.newSegFromArray('HNSHA', this.protoVersion === 300 ? 2 : 1, [1, NULL_1.NULL, [this.signIt.pin]]));
                }
                else {
                    this.addSeg(Helper_1.Helper.newSegFromArray('HNSHA', this.protoVersion === 300 ? 2 : 1, [1, NULL_1.NULL, [this.signIt.pin, this.signIt.tan]]));
                }
            }
            else {
                this.addSeg(Helper_1.Helper.newSegFromArray('HNSHA', 2, [2]));
            }
        }
        for (let i = 1; i !== this.segments.length; i += 1) {
            body += this.segments[i].transformForSend();
        }
        // Letztes segment erstellen
        if (this.signIt) {
            // in body ist der eigentliche body der dann aber jetzt neu erstellt wird
            // Verschlüsselung
            // 1. HNVSK                                     HNVSK:998:3
            // Sicherheitsprofil                            [PIN:1]
            // Sicherheitsfunktion, kodiert                 998 // bleibt immer so unabhängig von der der tatsächlichen Funktion
            // Rolle des Sicherheits-lieferanten, kodiert   1
            // Sicherheitsidentifikation, Details           [1.null.0]
            // Sicherheitsdatum und -uhrzeit                [1,20141216,205751]
            // Verschlüsselungs-algorithmus                 [2,2,13,@8@,5,1]
            // Schlüsselname                                [280:12345678:max:V:0:0]
            //      Ctry Code                               280 (hier fest)
            //      BLZ
            //      benutzer
            //      Schlüsselart                            V Chiffrierschlüssel
            //      Schlüsselnummer                         0
            //      Schlüsselversion                        0
            // Komprimierungsfunktion                       0
            // Zertifikat                                   leer hier
            // +998+1+1::0+1:20141216:205751+2:2:13:@8@:5:1+280:12345678:max:V:0:0+0'
            if (this.protoVersion === 300) {
                this.hnvsk = Helper_1.Helper.newSegFromArray('HNVSK', 3, [
                    ['PIN', this.signIt.pinVersion === '999' ? 1 : 2], 998, 1, [1, NULL_1.NULL, this.signIt.sysId],
                    [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                    [2, 2, 13, Helper_1.Helper.byte('\0\0\0\0\0\0\0\0'), 5, 1],
                    [280, this.signIt.blz, this.signIt.kundenId, 'V', 0, 0], 0,
                ]);
            }
            else {
                this.hnvsk = Helper_1.Helper.newSegFromArray('HNVSK', 2, [998, 1, [1, NULL_1.NULL, this.signIt.sysId],
                    [1, Helper_1.Helper.convertDateToDFormat(new Date()), Helper_1.Helper.convertDateToTFormat(new Date())],
                    [2, 2, 13, Helper_1.Helper.byte('\0\0\0\0\0\0\0\0'), 5, 1],
                    [280, this.signIt.blz, this.signIt.kundenId, 'V', 0, 0], 0,
                ]);
            }
            this.hnvsk.nr = 998;
            const segHNVSD = Helper_1.Helper.newSegFromArray('HNVSD', 1, [Helper_1.Helper.byte(body)]);
            segHNVSD.nr = 999;
            body = this.hnvsk.transformForSend();
            body += segHNVSD.transformForSend();
        }
        // Abschließen
        const seg = Helper_1.Helper.newSegFromArray('HNHBS', 1, [this.messageNumber]);
        this.addSeg(seg);
        body += seg.transformForSend();
        const llength = top.length + body.length;
        this.segments[0].store.data[0].data = Helper_1.Helper.getNrWithLeadingNulls(llength, 12);
        top = this.segments[0].transformForSend();
        return top + body;
    }
    addSeg(seg) {
        seg.nr = this.segmentsCounter + 1;
        this.segments[this.segmentsCounter] = seg;
        this.segmentsCounter += 1;
        return seg.nr;
    }
    isSigned() {
        return this.getSegmentsByName('HNSHK').length === 1;
    }
    getSegmentsByName(name) {
        return this.segments.filter(segment => segment.name === name);
    }
    getSegmentByReference(belongTo) {
        return this.segments.filter(segment => segment.referencedSegment === belongTo);
    }
    getSegmentsByNameAndReference(name, referencedSegment) {
        return this.segments.filter(segment => segment.name === name && segment.referencedSegment === referencedSegment);
    }
    getSegmentByNameAndReference(name, referencedSegment) {
        return this.segments.find(segment => segment.name === name && segment.referencedSegment === referencedSegment);
    }
    getSegmentByName(name) {
        return this.segments.find(segment => segment.name === name);
    }
    wasCanceled() {
        const statusSegments = this.getSegmentsByName(SegmentName_1.SegmentName.RETURN_STATUS_MESSAGE);
        const canceledSegment = statusSegments.find(statusSegment => {
            const canceledElement = statusSegment.store.data.find(dataElement => {
                return dataElement.data.getEl(1) === ReturnCode_1.ReturnCode.ERROR_CANCELED;
            });
            return canceledElement !== undefined;
        });
        return canceledSegment !== undefined;
    }
}
exports.Nachricht = Nachricht;
