"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Exceptions_1 = require("./Exceptions");
const Helper_1 = require("./Helper");
const Nachricht_1 = require("./Nachricht");
const NULL_1 = require("./NULL");
const OrderHelperChain_1 = require("./OrderHelperChain");
const SendMessage_1 = require("./SendMessage");
const SignInfo_1 = require("./SignInfo");
class Order {
    constructor(client) {
        this.client = client;
        this.error = null;
        this.intReqTan = false;
        this.intSendMsg = [];
        this.intGmsgList = [];
        this.checkKITypeAvailible = (kiType, vers, returnParam) => {
            if (kiType in this.client.bpd.gvParameters) {
                const pReturn = {};
                let testVers = [];
                if (vers instanceof Array) {
                    testVers = testVers.concat(vers);
                }
                else {
                    testVers.push(vers);
                }
                for (const vindex in testVers) {
                    if (testVers[vindex] in this.client.bpd.gvParameters[kiType]) {
                        if (returnParam) {
                            pReturn[vindex] = this.client.bpd.gvParameters[kiType][testVers[vindex]];
                        }
                        else {
                            return true;
                        }
                    }
                }
                if (returnParam) {
                    return pReturn;
                }
                return false;
            }
            if (returnParam) {
                return {};
            }
            return false;
        };
    }
    requireTan() {
        this.intReqTan = true;
    }
    msg(inData) {
        // 0. check no error
        if (this.error) {
            return false;
        }
        // 1. check if we support one of the segment versions
        let actVers = 0;
        if (inData.ki_type in this.client.bpd.gvParameters) {
            const availVers = Object.keys(inData.send_msg).sort((a, b) => {
                return (b - a);
            });
            for (const i in availVers) {
                if (availVers[i] in this.client.bpd.gvParameters[inData.ki_type]) {
                    actVers = availVers[i];
                    break;
                }
            }
        }
        if (actVers === 0) {
            this.error = new Exceptions_1.Exceptions.GVNotSupportedByKI(inData.ki_type, this.client.bpd.gvParameters[inData.ki_type]);
            return false;
        }
        // 2. Find the appropriate action
        let act = null;
        if (typeof inData.recv_msg === 'function') {
            act = inData.recv_msg;
        }
        else if (actVers in inData.recv_msg) {
            act = inData.recv_msg[actVers];
        }
        else if (0 in inData.recv_msg) {
            act = inData.recv_msg[0];
        }
        else {
            act = () => {
                // empty
            };
        }
        // 3. Prepare the Send Message object
        const sendMsg = new SendMessage_1.SendMessage();
        sendMsg.action = act;
        sendMsg.aufsetzpunktLoc = (inData.aufsetzpunkt_loc ? inData.aufsetzpunkt_loc : []);
        sendMsg.segment = Helper_1.Helper.newSegFromArray(inData.type, actVers, inData.send_msg[actVers]);
        sendMsg.version = actVers;
        this.intSendMsg.push(sendMsg);
    }
    done(cb) {
        // Exit CB is called when the function returns here it is checked if an error occures and then disconnects
        const extCb = (error, order, recvMsg) => __awaiter(this, void 0, void 0, function* () {
            if (error) {
                try {
                    yield this.client.close();
                    this.client.conLog.debug('Connection closed okay, after error.');
                }
                catch (closeError) {
                    this.client.conLog.error({
                        error: closeError,
                    }, 'Connection close failed after error.');
                }
            }
            cb(error, order, recvMsg);
        });
        // Main Part
        if (this.error) {
            extCb(this.error, this, null);
        }
        else {
            // Message prepare
            const perform = () => {
                const msg = new Nachricht_1.Nachricht(this.client.protoVersion);
                const signInfo = new SignInfo_1.SignInfo();
                signInfo.pin = this.client.pin;
                signInfo.tan = NULL_1.NULL;
                signInfo.sysId = this.client.sysId;
                signInfo.sigId = this.client.getNewSigId();
                signInfo.pinVersion = this.client.upd.availableTanVerfahren[0];
                msg.sign(signInfo);
                msg.init(this.client.dialogId, this.client.nextMsgNr, this.client.blz, this.client.kundenId);
                this.client.nextMsgNr += 1;
                // Fill in Segments
                for (const j in this.intSendMsg) {
                    const sendMessage = this.intSendMsg[j];
                    if (sendMessage.finished) {
                        continue;
                    }
                    // 1. Resolve Aufsetzpunkt if required, TODO here differentiate between versions
                    if (sendMessage.aufsetzpunkt) {
                        if (sendMessage.aufsetzpunktLoc.length >= 1) {
                            for (; sendMessage.segment.store.data.length < sendMessage.aufsetzpunktLoc[0];) {
                                sendMessage.segment.store.addDE(NULL_1.NULL);
                            }
                            if (sendMessage.aufsetzpunktLoc.length <= 1) {
                                // direkt
                                sendMessage.segment.store.data[sendMessage.aufsetzpunktLoc[0] - 1].data = sendMessage.aufsetzpunkt;
                            }
                            else {
                                // Unter DEG
                                extCb(new Exceptions_1.Exceptions.InternalError('Aufsetzpunkt Location is in DEG not supported yet.'), this, null);
                                return;
                            }
                        }
                        else {
                            extCb(new Exceptions_1.Exceptions.InternalError('Aufsetzpunkt Location is not set but an aufsetzpunkt was delivered'), this, null);
                            return;
                        }
                    }
                    // 2. Add Segment
                    msg.addSeg(sendMessage.segment);
                }
                // Send Segments to Destination
                this.client.sendMsgToDestination(msg, (error, recvMsg) => {
                    if (error) {
                        extCb(error, this, null);
                    }
                    else {
                        let gotAufsetzpunkt = false;
                        // 1. global Message testen
                        let gmsgException = null;
                        try {
                            const HIRMG = recvMsg.getSegmentsByName('HIRMG')[0];
                            for (const i in HIRMG.store.data) {
                                this.intGmsgList.push(HIRMG.store.data[i].data.data);
                                if (gmsgException == null && HIRMG.store.data[i].data.data[0].data.charAt(0) === '9') {
                                    gmsgException = new Exceptions_1.Exceptions.OrderFailedException(HIRMG.store.data[i].data.data);
                                }
                            }
                        }
                        catch (ee) {
                            extCb(new Exceptions_1.Exceptions.MalformedMessageFormat('HIRMG is mandatory but missing.'), this, recvMsg);
                            return;
                        }
                        if (gmsgException != null) {
                            extCb(gmsgException, this, recvMsg);
                            return;
                        }
                        // 2. einzelne Resp Segmente durchgehen
                        try {
                            for (const j in this.intSendMsg) {
                                const sendMessage = this.intSendMsg[j];
                                const relatedSegments = recvMsg.getSegmentByReference(sendMessage.segment.nr);
                                sendMessage.finished = true;
                                relatedSegments.forEach(segment => {
                                    if (segment.name !== 'HIRMS') {
                                        sendMessage.collectedSegments.push(segment);
                                        return;
                                    }
                                    const HIRMS = segment;
                                    HIRMS.store.data.forEach(deg => {
                                        sendMessage.collectedMessages.push(deg.data);
                                        if (deg.data.data[0].data === '3040') {
                                            // Got an Aufsetzpunkt
                                            try {
                                                sendMessage.aufsetzpunkt = deg.data.data[3].data;
                                            }
                                            catch (eee) {
                                                sendMessage.aufsetzpunkt = null;
                                            }
                                            sendMessage.finished = false;
                                            gotAufsetzpunkt = true;
                                        }
                                    });
                                });
                            }
                        }
                        catch (ee) {
                            extCb(new Exceptions_1.Exceptions.InternalError('Failed parsing Segments'), this, recvMsg);
                        }
                        // 3. check if we had an aufsetzpunkt
                        if (gotAufsetzpunkt) {
                            perform();
                        }
                        else {
                            // 4. Fertig die callbacks rufen
                            this.intSendMsg.forEach((sendMessage) => {
                                sendMessage.action(sendMessage.version, sendMessage.collectedSegments, sendMessage.collectedMessages, recvMsg);
                            });
                            extCb(null, this, recvMsg);
                        }
                    }
                });
            };
            perform();
        }
    }
    checkMessagesOkay(messages, throwError) {
        for (const i in messages) {
            const type = messages[i].data[0].data.charAt(0);
            if (type === '9') {
                if (throwError) {
                    throw new Exceptions_1.Exceptions.GVFailedAtKI(messages[i]);
                }
                return false;
            }
        }
        return true;
    }
    getSegByName(list, name) {
        const result = list.find(segment => segment.name === name);
        if (!result) {
            return null;
        }
        return result;
    }
    getElFromSeg(seg, nr, defaultValue) {
        if (!seg) {
            return defaultValue;
        }
        let e = null;
        try {
            e = seg.getEl(nr);
        }
        catch (e2) {
            e = defaultValue;
        }
        return e;
    }
    helper() {
        return new OrderHelperChain_1.OrderHelperChain(this);
    }
}
exports.Order = Order;
