"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class ParseError {
    constructor(location, msg, pos) {
        this.location = location;
        this.msg = msg;
        this.pos = pos;
    }
    toString() {
        return this.msg;
    }
}
exports.ParseError = ParseError;
class Parser {
    constructor(data) {
        this.data = data;
        this.curPos = 0;
        this.marker = {};
    }
    getCurPos() {
        return this.curPos;
    }
    getData() {
        return this.data;
    }
    setMarker(mark, pos) {
        this.marker[mark] = pos;
    }
    setMarkerWithCurrentPos(mark) {
        this.setMarker(mark, this.curPos);
    }
    setPosBackToMarker(mark) {
        this.curPos = this.marker[mark];
    }
    getCurrentPos() {
        return this.curPos;
    }
    setCurrentPos(pos) {
        this.curPos = pos;
    }
    getCurrentChar() {
        return this.data[this.curPos];
    }
    hasNext() {
        return this.curPos < this.data.length;
    }
    nextPos() {
        if (!this.hasNext()) {
            return false;
        }
        this.curPos += 1;
        return true;
    }
    gotoPos(pos) {
        if (pos === -1) {
            this.curPos = this.data.length;
            return false;
        }
        this.curPos = pos;
        return true;
    }
    getTextFromMarkerToCurrentPos(mark) {
        return this.getTextFromPostoPos(this.marker[mark], this.curPos);
    }
    getTextFromPostoPos(posFrom, posTo) {
        return this.data.substr(posFrom, posTo - posFrom);
    }
    findNextValidChar(validChars) {
        for (let i = this.curPos; i < this.data.length; i += 1) {
            if (validChars.includes(this.data[i])) {
                return i;
            }
        }
        return -1;
    }
    gotoNextValidChar(validChars) {
        return this.gotoPos(this.findNextValidChar(validChars));
    }
    // This goes to the first char of the string
    findNextValidString(validStrings) {
        const origPos = this.curPos;
        let validChars = '';
        validStrings.forEach(s => {
            validChars += s.charAt(0);
        });
        let pos = this.curPos;
        do {
            pos = this.findNextValidChar(validChars);
            if (pos !== -1) {
                for (let i = 0; i !== validStrings.length; i += 1) {
                    if (validStrings[i].charAt(0) === this.data[pos]) {
                        // prüfen ob voll passt
                        const compStr = this.data.substr(pos, validStrings[i].length);
                        if (compStr === validStrings[i]) {
                            this.curPos = origPos;
                            return pos;
                        }
                    }
                }
                this.curPos = pos + 1;
            }
        } while (pos !== -1);
        this.curPos = origPos;
        return pos;
    }
    gotoNextValidString(validStrings) {
        return this.gotoPos(this.findNextValidString(validStrings));
    }
    gotoNextValidCharButIgnoreWith(validChars, demask) {
        while (true) {
            const pos = this.findNextValidChar(validChars);
            if (pos === -1) {
                this.curPos = this.data.length;
                return false;
            }
            if (pos === 0) {
                this.curPos = pos;
                return true;
            }
            if (demask.includes(this.data[pos - 1])) {
                if ((pos + 1) >= this.data.length) {
                    this.curPos = pos;
                    return false;
                }
                this.curPos = pos + 1;
                // retry
            }
            else {
                this.curPos = pos;
                return true;
            }
        }
    }
}
exports.Parser = Parser;
