"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DatenElementGruppe_1 = require("./DatenElementGruppe");
const NULL_1 = require("./NULL");
const Parser_1 = require("./Parser");
class Segment {
    constructor() {
        this.store = new DatenElementGruppe_1.DatenElementGruppe();
        this.name = null;
        this.referencedSegment = 0;
    }
    init(name, nr, version, referencedSegment) {
        this.name = name;
        this.nr = nr;
        this.version = version;
        this.referencedSegment = referencedSegment;
    }
    transformForSend() {
        let result = '';
        result += this.name; // Nr. 1 Segmentkennung an ..6 M 1
        result += ':' + this.nr; // Nr. 2 Segmentnummer num ..3 M 1 >=1
        result += ':' + this.version; // Nr. 3 Segmentversion GD num ..3 M 1
        if (this.referencedSegment !== 0)
            result += ':' + this.referencedSegment;
        for (let i = 0; i !== this.store.data.length; i += 1) {
            if (this.store.data[i].data !== NULL_1.NULL) {
                if (this.store.data[i].desc === 1) {
                    result += '+' + this.store.data[i].data; // DE
                }
                else if (this.store.data[i].desc === 2) {
                    const degData = this.store.data[i].data;
                    result += '+' + degData.transformForSend(); // DEG
                }
                else if (this.store.data[i].desc === 3) {
                    const binData = this.store.data[i].data;
                    result += '+@' + binData.length + '@' + binData; // BIN DAT
                }
            }
            else {
                // leer
                result += '+';
            }
        }
        result += "'";
        return result;
    }
    parse(parser) {
        let startPos = parser.getCurrentPos();
        // 1. Segmentkopf
        // Nr. 1 Segmentkennung an ..6 M 1
        parser.setMarkerWithCurrentPos('start');
        if (parser.gotoNextValidChar(':')) {
            this.name = parser.getTextFromMarkerToCurrentPos('start');
        }
        else {
            throw new Parser_1.ParseError('Seg', 'Segmentkennung Fehlt!', startPos);
        }
        // Nr. 2 Segmentnummer num ..3 M 1 >=1
        parser.nextPos();
        startPos = parser.getCurrentPos();
        parser.setMarkerWithCurrentPos('start');
        if (parser.gotoNextValidChar(':')) {
            this.nr = parseInt(parser.getTextFromMarkerToCurrentPos('start'), 10);
        }
        else {
            throw new Parser_1.ParseError('Seg', 'Segmentnummer fehlt!', startPos);
        }
        // Nr. 3 Segmentversion GD num ..3 M 1
        parser.nextPos();
        startPos = parser.getCurrentPos();
        parser.setMarkerWithCurrentPos('start');
        if (parser.gotoNextValidChar(":+'")) {
            this.version = parser.getTextFromMarkerToCurrentPos('start');
        }
        else {
            throw new Parser_1.ParseError('Seg', 'Segmentversion fehlt!', startPos);
        }
        // Nr. 4 Bezugssegment GD num ..3 K 1 >=1
        if (parser.getCurrentChar() === ':') {
            parser.nextPos();
            startPos = parser.getCurrentPos();
            parser.setMarkerWithCurrentPos('start');
            if (parser.gotoNextValidChar('+')) {
                this.referencedSegment = parseInt(parser.getTextFromMarkerToCurrentPos('start'), 10);
            }
            else {
                throw new Parser_1.ParseError('Seg', 'Unerwartetes ENDE!', startPos);
            }
        }
        // jetzt kommen datenlemente oder datenelementgruppen
        while (parser.getCurrentChar() !== "'" && parser.hasNext()) {
            parser.nextPos();
            startPos = parser.getCurrentPos();
            parser.setMarkerWithCurrentPos('start');
            if (parser.getCurrentChar() === '@') {
                // binary
                parser.nextPos();
                parser.setMarkerWithCurrentPos('start');
                if (!parser.gotoNextValidChar('@'))
                    throw new Parser_1.ParseError('Seg', 'Error binary!', startPos);
                const len = parseInt(parser.getTextFromMarkerToCurrentPos('start'), 10);
                parser.nextPos();
                parser.setMarkerWithCurrentPos('start');
                parser.setCurrentPos(parser.getCurrentPos() + len);
                if ("+:'".indexOf(parser.getCurrentChar()) === -1)
                    throw new Parser_1.ParseError('Seg', 'Error binary, Wrong Length!' + len, startPos);
                this.store.addDEbin(parser.getTextFromMarkerToCurrentPos('start'));
            }
            else if (parser.gotoNextValidCharButIgnoreWith("+:'", '?')) {
                if (parser.getCurrentChar() === '+' || parser.getCurrentChar() === "'") {
                    // Normales datenelement
                    this.store.addDE(parser.getTextFromMarkerToCurrentPos('start'));
                }
                else {
                    // Datengruppe
                    parser.setPosBackToMarker('start');
                    const neuDeg = new DatenElementGruppe_1.DatenElementGruppe();
                    neuDeg.parse(parser);
                    this.store.addDEG(neuDeg);
                }
            }
            else {
                throw new Parser_1.ParseError('Seg', 'Unerwartetes ENDE!', startPos);
            }
        }
    }
    getEl(nr) {
        return this.store.data[nr - 1];
    }
    getElString(nr) {
        return this.store.data[nr - 1].data;
    }
}
exports.Segment = Segment;
